                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--

*/

#include <afxwin.h>
#include <afxtempl.h>

#include <d3d8.h>
#include <d3dx8.h>

D3DCAPS8 caps;
LPDIRECT3D8       d3d       = NULL;
LPDIRECT3DDEVICE8 d3ddevice = NULL;

#include "DirectX/GLizer.h"
GLizer glizer;

#include "Common/CPU.h"

#include "DirectX/Texture.h"
#include "3D/Model.h"
#include "3D/Resolution.h"

#define VK_ENTER 13


/* ################################################################################################################### */

class Direct3DWindow : public CFrameWnd {
public:
    Direct3DWindow();

	void DrawFrame();
protected:
	TextureHandler *texHandler;
	TextureID text, tex0, tex1;

	CRect windowedRect;
	int Width, Height;
	int Refresh;
	bool vSync;
	bool fullScreen;
	bool captureMouse;
	bool windowMoved;
	bool showFPS;
	int middleX, middleY;

	bool Key[256];
	bool leftButton;
	float mouseSensibility;
	bool invertMouse;
	int mouseX, mouseY;

    __int64 cpuHz;

	Vertex position;
	float Wx, Wy, Wz;


	Model *model;


	BOOL PreCreateWindow(CREATESTRUCT &cs);
	int  OnCreate(LPCREATESTRUCT lpCreateStruct);
	void OnSize(UINT nType, int cx, int cy);
	void OnClose();

	void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	void OnKeyUp  (UINT nChar, UINT nRepCnt, UINT nFlags);
	void OnSysKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);

	void OnLButtonDown(UINT nFlags, CPoint point);
	void OnLButtonUp  (UINT nFlags, CPoint point);
	void OnMouseMove  (UINT nFlags, CPoint point);

	void OnKillFocus(CWnd *pNewWnd);

	DECLARE_MESSAGE_MAP ()
};

BEGIN_MESSAGE_MAP(Direct3DWindow, CFrameWnd)
	
	ON_WM_CREATE ()
	ON_WM_CLOSE ()

	ON_WM_KEYDOWN ()
	ON_WM_KEYUP ()
	ON_WM_SYSKEYDOWN ()

	ON_WM_LBUTTONDOWN ()
	ON_WM_LBUTTONUP ()
	ON_WM_MOUSEMOVE ()

	ON_WM_KILLFOCUS ()

END_MESSAGE_MAP()


/* ################################################################################################################### */

void Direct3DWindow::DrawFrame(){
    static bool first = true;

    static __int64 prev, curr;
    static float currTime;
    float time, speed;

    prev = curr;
    curr = getCycleNumber();
    if (first){
        prev = curr;
        first = false;
    }
    time = (float) (100000.0 * double(curr - prev) / cpuHz);
    currTime += time;
    speed = time * 0.005f;



	D3DXMATRIX modelView, projection, tmp;
	
	D3DXMatrixIdentity(&tmp);
	d3ddevice->SetTransform(D3DTS_WORLD, &tmp);
	
	D3DXMatrixRotationZ(&modelView, -Wz * PI / 180);
	D3DXMatrixRotationX(&tmp, -Wx * PI / 180);
	D3DXMatrixMultiply(&modelView, &tmp, &modelView);
	D3DXMatrixRotationY(&tmp, -Wy * PI / 180);
	D3DXMatrixMultiply(&modelView, &tmp, &modelView);
	D3DXMatrixTranslation(&tmp, -position.x, -position.y, -position.z);
	D3DXMatrixMultiply(&modelView, &tmp, &modelView);
	d3ddevice->SetTransform(D3DTS_VIEW, &modelView);

	D3DXMatrixPerspectiveFovLH(&projection, PI / 4, (float) Width / (float) Height, 1, 32000);
	d3ddevice->SetTransform(D3DTS_PROJECTION, &projection);



	Vertex dx(modelView[0], modelView[4], modelView[8]);
	Vertex dy(modelView[1], modelView[5], modelView[9]);
	Vertex dz(modelView[2], modelView[6], modelView[10]);

	Vertex dir(0,0,0);

	if (Key[VK_LEFT ]) dir -= dx;
	if (Key[VK_RIGHT]) dir += dx;

	if (Key[VK_DOWN]) dir -= dz;
	if (Key[VK_UP  ]) dir += dz;

	if (Key[VK_SHIFT  ]) dir -= dy;
	if (Key[VK_CONTROL]) dir += dy;

	if (Key[VK_ENTER]){
		position = Vertex(200,200,200);
		Wx = 40;
		Wy = 225;
		Wz = 0;
	}

	if (lengthSqr(dir) != 0){
		dir.normalize();
	}


	Vertex oldPos = position;
	position += speed * dir;



    d3ddevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    d3ddevice->BeginScene();

	
		float x,y,z;

		x = 128 * sinf(currTime * 0.00002f);
		y = 128 * cosf(currTime * 0.00002f);
		z = 128 * (1 + cosf(currTime * 0.0000354f) * sinf(currTime * 0.00001152f));
		

		D3DLIGHT8 light;
		ZeroMemory(&light, sizeof(light));
		light.Type = D3DLIGHT_POINT;
		light.Diffuse.r = 1.0f;
		light.Diffuse.g = 1.0f;
		light.Diffuse.b = 1.0f;
		light.Position  = D3DXVECTOR3(x,y,z);
		light.Range = 32000;
		light.Attenuation0 = 1.0f;
		light.Attenuation1 = 0.0f;
		light.Attenuation2 = 0.00002f;
		light.Ambient.r = 1.0f;
		light.Ambient.g = 1.0f;
		light.Ambient.b = 1.0f;
		light.Specular.r = 1.0f;
		light.Specular.g = 1.0f;
		light.Specular.b = 1.0f;


		d3ddevice->SetLight(0, &light);
		
		d3ddevice->LightEnable(0, TRUE);
	    d3ddevice->SetRenderState(D3DRS_LIGHTING,       TRUE);
	    d3ddevice->SetRenderState(D3DRS_SPECULARENABLE, TRUE);

		D3DMATERIAL8 mtrl;
		ZeroMemory(&mtrl, sizeof(mtrl));
		mtrl.Diffuse.r = 1.0f;
		mtrl.Diffuse.g = 1.0f;
		mtrl.Diffuse.b = 1.0f;
		mtrl.Ambient.r = 0.2f;
		mtrl.Ambient.g = 0.2f;
		mtrl.Ambient.b = 0.2f;
		mtrl.Specular.r = 0.4f;
		mtrl.Specular.g = 0.4f;
		mtrl.Specular.b = 0.4f;
		mtrl.Power = 16;
		d3ddevice->SetMaterial(&mtrl);

		static float cTime;
		cTime += time;

		if (cTime < 300000.0f){
			model->draw(0, 1, 0.5f * (1 + sinf(cTime * 0.00002f * PI - PI * 0.5f)), texHandler);
		} else if (cTime < 400000.0f){
			model->draw(0, 2, 0.5f * (1 - sinf(cTime * 0.00001f * PI - PI * 0.5f)), texHandler);
		} else if (cTime < 500000.0f){
			model->draw(0, 2, 1 - 0.1f * (1 + sinf(cTime * 0.00002f * PI - PI * 0.5f)), texHandler);
		} else if (cTime < 600000.0f){
			model->draw(0, 2, 0.5f * (1 + sinf(cTime * 0.00001f * PI - PI * 0.5f)), texHandler);
		} else if (cTime < 800000.0f){
			model->draw(0, 1, 0.5f * (1 + sinf(cTime * 0.00002f * PI - PI * 0.5f)), texHandler);
		} else if (cTime < 900000.0f){
			model->draw(0, 3, 0.5f * (1 + sinf(cTime * 0.00001f * PI - PI * 0.5f)), texHandler);
		} else if (cTime < 1000000.0f){
			model->draw(0, 3, 1 - 0.1f * (1 + sinf(cTime * 0.00002f * PI - PI * 0.5f)), texHandler);
		} else if (cTime < 1100000.0f){
			model->draw(0, 3, 0.5f * (1 - sinf(cTime * 0.00001f * PI - PI * 0.5f)), texHandler);
		} else {
			cTime = 0;
		}

		d3ddevice->SetRenderState(D3DRS_SPECULARENABLE, FALSE);
	    d3ddevice->SetRenderState(D3DRS_LIGHTING,       FALSE);
		d3ddevice->LightEnable(0, FALSE);


		texHandler->setTextures(tex1);
		texHandler->setBlending(D3DBLEND_ONE, D3DBLEND_ONE);
		texHandler->apply();

		Vertex pos(x,y,z);

		glizer.glBegin(D3DPT_TRIANGLEFAN, false, false, 1);
			glizer.glTexCoord(0, 0);
			glizer.glVertex(pos + 25 * (-dx - dy));

			glizer.glTexCoord(1, 0);
			glizer.glVertex(pos + 25 * ( dx - dy));

			glizer.glTexCoord(1, 1);
			glizer.glVertex(pos + 25 * ( dx + dy));

			glizer.glTexCoord(0, 1);
			glizer.glVertex(pos + 25 * (-dx + dy));

		glizer.glEnd();


		if (showFPS){
			static float fps[8];
			char str[256];
			int i;
			
			for (i = 0; i < 7; i++){
				fps[i] = fps[i+1];
			}
			
			fps[7] = 100000 / time;
			float sum = 0;
			for (i = 0; i < 8; i++) sum += fps[i];
			
			sprintf(str,"%d", (int) (sum / 8 + 0.5f));
			
			texHandler->setTextures(text);
			texHandler->setTexEnvs(TEXENV_REPLACE);
			texHandler->setBlending(D3DBLEND_SRCALPHA, D3DBLEND_INVSRCALPHA);
			texHandler->apply();
			
			d3dDrawText(str, 0.02f, 0.02f, 0.03f, 0.03f);
		}
		


    d3ddevice->EndScene();






    d3ddevice->Present(NULL, NULL, NULL, NULL);
}


Direct3DWindow::Direct3DWindow(){
	Width   = 1024;
	Height  = 768;
	Refresh = 85;
	vSync = false;
	fullScreen = true;
	captureMouse = true;

	mouseSensibility = 0.1f;
	invertMouse = true;

	middleX = Width / 2;
	middleY = Height / 2;

    SetCursorPos(middleX, middleY);

	showFPS = true;

	position = Vertex(200,200,200);
	Wx = 40;
	Wy = 225;
	Wz = 0;

	ZeroMemory(&Key, sizeof(Key));
	leftButton = false;

	CRect r(0, 0, Width, Height);
	windowedRect = r;

	Create(NULL, "Direct3D", WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS, r);
}

BOOL Direct3DWindow::PreCreateWindow(CREATESTRUCT &cs){
	CFrameWnd::PreCreateWindow(cs);

	cs.dwExStyle &= ~WS_EX_CLIENTEDGE;

	return 1;
}


int Direct3DWindow::OnCreate(LPCREATESTRUCT lpCreateStruct){
	if (CFrameWnd::OnCreate(lpCreateStruct) == -1) return -1;

	ShowWindow(SW_SHOW);
	SetForegroundWindow();
	SetFocus();

    initCPU();
    cpuHz = getHz();

	texHandler = NULL;
	if ((d3d = Direct3DCreate8(D3D_SDK_VERSION)) == NULL){
		MessageBox("Couldn't initialize Direct3D\nMake sure you have installed DirectX 8 or later.", "Error", MB_OK | MB_ICONERROR);
		PostMessage(WM_CLOSE);
		return 0;
	}
	

	d3d->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &caps);

	if (!(caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT) &&
		 (caps.VertexProcessingCaps & D3DVTXPCAPS_TWEENING)){
		MessageBox("No hardware T&L with tweening support found.", "Error", MB_OK | MB_ICONERROR);
		PostMessage(WM_CLOSE);
		return 0;
	}

	if (caps.MaxStreams < 3){
		CString str;
		str.Format("Your card supports only %d vertex streams, you need at least 3 to run this demo.", caps.MaxStreams);
		MessageBox(str, "Error", MB_OK | MB_ICONERROR);
		PostMessage(WM_CLOSE);
		return 0;
	}

    D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory(&d3dpp, sizeof(d3dpp));

	if (fullScreen){
		D3DFORMAT formats[] = {D3DFMT_A8R8G8B8, D3DFMT_X8R8G8B8};
		if (!d3dFindBestResolution(Width, Height, formats, sizeof(formats) / sizeof(D3DFORMAT), Refresh)){
			MessageBox("No suitable displaymode found.", "Error", MB_OK | MB_ICONERROR);
			PostMessage(WM_CLOSE);
			return 0;
		}

		d3dpp.Windowed = FALSE;
		d3dpp.BackBufferWidth  = Width;
		d3dpp.BackBufferHeight = Height;
		d3dpp.BackBufferFormat = formats[0];
		
		d3dpp.FullScreen_RefreshRateInHz = Refresh;
		if (vSync){
			d3dpp.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_ONE;
			d3dpp.BackBufferCount = 2;
		} else {
			d3dpp.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
			d3dpp.BackBufferCount = 1;
		}
	} else {
		D3DDISPLAYMODE displayMode;
		d3d->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &displayMode);

		d3dpp.Windowed = TRUE;
		d3dpp.BackBufferWidth  = Width;
		d3dpp.BackBufferHeight = Height;
		d3dpp.BackBufferFormat = displayMode.Format;
		d3dpp.BackBufferCount  = 1;
		
	}
	d3dpp.EnableAutoDepthStencil = TRUE;
	d3dpp.AutoDepthStencilFormat = D3DFMT_D24S8;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;

	checkTextureFormats(d3dpp.BackBufferFormat);
	
    if (FAILED(d3d->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, (HWND) *this, D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &d3ddevice))){
		MessageBox("Couldn't create Direct3D device interface.", "Error", MB_OK | MB_ICONERROR);
		PostMessage(WM_CLOSE);
        return 0;
    }
	


	// Disable some odd defaults
	d3ddevice->SetRenderState(D3DRS_LIGHTING, FALSE);
	d3ddevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);


	texHandler = new TextureHandler();
	text     = texHandler->loadFromFile("..\\Textures\\font2.dds");

	tex0     = texHandler->loadFromFile("Bub.png");
	tex1     = texHandler->loadFromFile("..\\Textures\\Corona.png");

	texHandler->uploadAll();

	model = new Model();
	model->setTexture(tex0);
	model->loadFromFile("bub.model");

	
	ShowCursor(FALSE);

	return 0;
}

void Direct3DWindow::OnSize(UINT nType, int cx, int cy){
	D3DVIEWPORT8 vp;
	vp.X      = 0;
	vp.Y      = 0;
	vp.Width  = cx;
	vp.Height = cy;
	vp.MinZ   = 0.0f;
	vp.MaxZ   = 1.0f;
	d3ddevice->SetViewport(&vp);
}

void Direct3DWindow::OnClose(){
	if (texHandler != NULL){
		delete model;

		texHandler->unloadAll();
		delete texHandler;
	}

    if (d3ddevice != NULL) d3ddevice->Release();
    if (d3d != NULL) d3d->Release();

	ShowCursor(TRUE);
	DestroyWindow();
}


void Direct3DWindow::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags){
	Key[nChar & 0xFF] = true;

	switch(nChar){
	case ' ':
		showFPS = !showFPS;
		break;
	case VK_ESCAPE:
		if (!fullScreen && captureMouse){
			ShowCursor(TRUE);
			captureMouse = false;
		} else {
			PostMessage(WM_CLOSE, 0,0);
		}
		break;
	}
}

void Direct3DWindow::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags){
	Key[nChar & 0xFF] = false;
}

void Direct3DWindow::OnSysKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags){
	if (((nFlags >> 13) & 0x3) == 1 && nChar == VK_ENTER){
		if (fullScreen){
			//resetResolution();
			middleX = (windowedRect.left + windowedRect.right)  / 2;
			middleY = (windowedRect.top  + windowedRect.bottom) / 2;
			MoveWindow(windowedRect, TRUE);

			fullScreen = false;
		} else {
			middleX = Width  / 2;
			middleY = Height / 2;
			MoveWindow(0, 0, Width, Height, TRUE);
			//setResolution(Width, Height, Bpp, Refresh);

			captureMouse = true;
			ShowCursor(FALSE);

			fullScreen = true;
		}
	}
}

void Direct3DWindow::OnLButtonDown(UINT nFlags, CPoint point){
	mouseX = point.x;
	mouseY = point.y;

	windowMoved = false;
	leftButton = true;
}

void Direct3DWindow::OnLButtonUp(UINT nFlags, CPoint point){
	if (!captureMouse){
		if (!windowMoved){
			captureMouse = true;
			ShowCursor(FALSE);
		}
	}

	leftButton = false;
}

void Direct3DWindow::OnMouseMove(UINT nFlags, CPoint point){
	static bool changeMouse = false;

	if (captureMouse){
		ClientToScreen(&point);

		Wx += mouseSensibility * (point.y - middleY) * (invertMouse? -1 : 1);
		Wy += mouseSensibility * (point.x - middleX);
		
		if (changeMouse = !changeMouse) SetCursorPos(middleX, middleY);
	} else {
		if (leftButton){
			windowedRect += CSize(point.x - mouseX, point.y - mouseY);
			middleX = (windowedRect.left + windowedRect.right)  / 2;
			middleY = (windowedRect.top  + windowedRect.bottom) / 2;

			MoveWindow(windowedRect, TRUE);
			windowMoved = true;
		}
	}
}

void Direct3DWindow::OnKillFocus(CWnd *pNewWnd){
	if (fullScreen){
		//resetResolution();
		middleX = (windowedRect.left + windowedRect.right)  / 2;
		middleY = (windowedRect.top  + windowedRect.bottom) / 2;
		MoveWindow(windowedRect, TRUE);
		fullScreen = false;
		captureMouse = false;
		ShowCursor(TRUE);
	}
}

/* ################################################################################################################### */


class Direct3DApp : public CWinApp {
public:
    virtual BOOL InitInstance();
	BOOL OnIdle(LONG lCount);
};


BOOL Direct3DApp::InitInstance(){
	m_pMainWnd = new Direct3DWindow();

	return TRUE;
}

BOOL Direct3DApp::OnIdle(LONG lCount){
	
	((Direct3DWindow *) m_pMainWnd)->DrawFrame();

	return TRUE;
}

Direct3DApp Direct3DApp;


/* ########################################## End of file ############################################################ */
